# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import collections
import glob
import os
import sys
from pathlib import Path
from typing import TypedDict

import pmb.chroot
import pmb.helpers.cli
from pmb.core import Chroot
from pmb.helpers import logging
from pmb.types import Env, RunOutputTypeDefault


class CiScriptDescriptor(TypedDict):
    description: str
    options: list[str]
    artifacts: str | None


def get_ci_scripts(topdir: Path) -> dict[str, CiScriptDescriptor]:
    """
    Find 'pmbootstrap ci'-compatible scripts inside a git repository, and
    parse their metadata (description, options). The reference is at:
    https://postmarketos.org/pmb-ci

    :param topdir: top directory of the git repository, get it with: pmb.helpers.git.get_topdir()

    :returns: a dict of CI scripts found in the git repository, e.g.
      {"ruff": {"description": "lint all python scripts", "options": []}, ...}
    """
    ret: dict[str, CiScriptDescriptor] = {}
    for script in glob.glob(f"{topdir}/.ci/*.sh"):
        is_pmb_ci_script = False
        description = ""
        options = []
        artifacts = None

        with open(script) as handle:
            for line in handle:
                if line.startswith("# https://postmarketos.org/pmb-ci"):
                    is_pmb_ci_script = True
                elif line.startswith("# Description: "):
                    description = line.split(": ", 1)[1].rstrip()
                elif line.startswith("# Options: "):
                    options = line.split(": ", 1)[1].rstrip().split(" ")
                elif line.startswith("# Artifacts: "):
                    artifacts = line.split(": ", 1)[1].strip()
                elif not line.startswith("#"):
                    # Stop parsing after the block of comments on top
                    break

        if not is_pmb_ci_script:
            continue

        if not description:
            logging.error(f"ERROR: {script}: missing '# Description: …' line")
            sys.exit(1)

        for option in options:
            if option not in pmb.config.ci_valid_options:
                raise RuntimeError(
                    f"{script}: unsupported option '{option}'."
                    " Typo in script or pmbootstrap too old?"
                )

        short_name = os.path.basename(script).split(".", -1)[0]
        ret[short_name] = {"description": description, "options": options, "artifacts": artifacts}
    return ret


def sort_scripts_by_speed(scripts: dict[str, CiScriptDescriptor]) -> dict[str, CiScriptDescriptor]:
    """
    Order the scripts, so fast scripts run before slow scripts. Whether a
    script is fast or not is determined by the '# Options: slow' comment in
    the file.

    :param scripts: return of get_ci_scripts()

    :returns: same format as get_ci_scripts(), but as ordered dict with
      fast scripts before slow scripts

    """
    ret = collections.OrderedDict()

    # Fast scripts first
    for script_name, script in scripts.items():
        if "slow" in script["options"]:
            continue
        ret[script_name] = script

    # Then slow scripts
    for script_name, script in scripts.items():
        if "slow" not in script["options"]:
            continue
        ret[script_name] = script
    return ret


def ask_which_scripts_to_run(
    scripts_available: dict[str, CiScriptDescriptor],
) -> dict[str, CiScriptDescriptor]:
    """
    Display an interactive prompt about which of the scripts the user
    wishes to run, or all of them.

    :param scripts_available: same format as get_ci_scripts()

    :returns: either full scripts_available (all selected), or a subset

    """
    count = len(scripts_available.items())
    choices = []

    logging.info(f"Available CI scripts ({count}):")
    for script_name, script in scripts_available.items():
        extra = ""
        if "slow" in script["options"]:
            extra += " (slow)"
        logging.info(f"* {script_name}: {script['description']}{extra}")
        choices += [script_name]

    choice_regex = "|".join(choices)
    ci_regex = f"(all)|({choice_regex})(,({choice_regex}))*"
    choices += ["all"]
    logging.info('Fill a comma-separated list of scripts or "all"')
    selection = pmb.helpers.cli.ask(
        "Scripts", None, "all", validation_regex=ci_regex, complete=choices
    )
    if selection == "all":
        return scripts_available
    else:
        return {script: scripts_available[script] for script in selection.split(",")}


def copy_git_repo_to_chroot(topdir: Path, include_dot_git_dir: bool = False) -> None:
    """
    Create a tarball of the git repo (including unstaged changes and new
    files) and extract it in chroot_native.

    :param topdir: top directory of the git repository, get it with:
      pmb.helpers.git.get_topdir()
    :param include_dot_git_dir: Also include .git
    """
    chroot = Chroot.native()
    pmb.chroot.init(chroot)
    tarball_path = chroot / "tmp/git.tar.gz"
    files = pmb.helpers.git.get_files(topdir, include_dot_git_dir=include_dot_git_dir)

    with open(f"{tarball_path}.files", "w") as handle:
        for file in files:
            handle.write(file)
            handle.write("\n")

    pmb.helpers.run.user(["tar", "-cf", tarball_path, "-T", f"{tarball_path}.files"], topdir)

    ci_dir = Path("/home/pmos/ci")
    pmb.chroot.user(["rm", "-rf", ci_dir])
    pmb.chroot.user(["mkdir", ci_dir])
    pmb.chroot.user(["tar", "-xf", "/tmp/git.tar.gz"], working_dir=ci_dir)


def run_scripts(topdir: Path, scripts: dict[str, CiScriptDescriptor]) -> None:
    """
    Run one of the given scripts after another, either natively or in a
    chroot. Display a progress message and stop on error (without printing
    a python stack trace).

    :param topdir: top directory of the git repository, get it with:
      pmb.helpers.git.get_topdir()

    :param scripts: return of get_ci_scripts()

    """
    steps = len(scripts)
    step = 0
    repo_copied = False

    for script_name, script in scripts.items():
        step += 1

        where = "pmbootstrap chroot"
        if "native" in script["options"]:
            where = "native"

        script_path = f".ci/{script_name}.sh"
        logging.info(f"*** ({step}/{steps}) RUNNING CI SCRIPT: {script_path} [{where}] ***")

        if "native" in script["options"]:
            rc = pmb.helpers.run.user([script_path], topdir, output=RunOutputTypeDefault.TUI)
            continue
        else:
            # Run inside pmbootstrap chroot
            if not repo_copied:
                copy_git_repo_to_chroot(
                    topdir, include_dot_git_dir="with-dot-git" in script["options"]
                )
                repo_copied = True

            env: Env = {"TESTUSER": "pmos"}
            rc = pmb.chroot.root(
                [script_path],
                check=False,
                env=env,
                working_dir=Path("/home/pmos/ci"),
                output=RunOutputTypeDefault.TUI,
            )
        if rc:
            logging.error(f"ERROR: CI script failed: {script_name}")
            sys.exit(1)

        if script["artifacts"]:
            logging.info(f"Copy CI artifacts to ./ci-artifacts/{script_name}")
            path = Chroot.native().path / "home/pmos/ci" / script["artifacts"]
            target = topdir / "ci-artifacts" / script_name
            pmb.helpers.run.user(["rm", "-rf", target])
            pmb.helpers.run.user(["mkdir", "-p", target])
            pmb.helpers.run.user(["cp", "-r", path, target])



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import collections
import tarfile
from collections.abc import Sequence
from pathlib import Path
from typing import Any, Literal, cast, overload

import pmb.helpers.package
import pmb.helpers.repo
import pmb.parse.version
from pmb.core.apkindex_block import ApkindexBlock
from pmb.core.arch import Arch
from pmb.helpers import logging

apkindex_map = {
    "A": "arch",
    "D": "depends",
    "o": "origin",
    "P": "pkgname",
    "p": "provides",
    "k": "provider_priority",
    "t": "timestamp",
    "V": "version",
}

required_apkindex_keys = ["arch", "pkgname", "version"]


def parse_next_block(path: Path, lines: list[str]) -> ApkindexBlock | None:
    """
    Parse the next block in an APKINDEX.

    :param path: to the APKINDEX.tar.gz
    :param start: current index in lines, gets increased in this
                  function. Wrapped into a list, so it can be modified
                  "by reference". Example: [5]
    :param lines: all lines from the "APKINDEX" file inside the archive
    :returns: ApkindexBlock
    :returns: None, when there are no more blocks
    """
    # Parse until we hit an empty line or end of file
    ret: dict[str, Any] = {}
    required_found = 0  # Count the required keys we found
    line = ""
    while len(lines):
        # We parse backwards for performance (pop(0) is super slow)
        line = lines.pop()
        if not line:
            continue
        # Parse keys from the mapping
        k = line[0]
        key = apkindex_map.get(k)

        # The checksum key is always the FIRST in the block, so when we find
        # it we know we're done.
        if k == "C":
            break
        if key:
            if key in ret:
                raise RuntimeError(f"Key {key} specified twice in block: {ret}, file: {path}")
            if key in required_apkindex_keys:
                required_found += 1
            ret[key] = line[2:]

    # Format and return the block
    if not len(lines) and not ret:
        return None

    # Check for required keys
    if required_found != len(required_apkindex_keys):
        for key in required_apkindex_keys:
            if key not in ret:
                raise RuntimeError(f"Missing required key '{key}' in block {ret}, file: {path}")
        raise RuntimeError(
            f"Expected {len(required_apkindex_keys)} required keys,"
            f" but found {required_found} in block: {ret}, file: {path}"
        )

    # Format optional lists
    for key in ["provides", "depends"]:
        if key in ret and ret[key] != "":
            # Ignore all operators for now
            values = ret[key].split(" ")
            ret[key] = []
            for value in values:
                for operator in [">", "=", "<", "~"]:
                    if operator in value:
                        value = value.split(operator)[0]
                        break
                ret[key].append(value)
        else:
            ret[key] = []
    provider_priority = ret.get("provider_priority")
    if provider_priority:
        if not provider_priority.isdigit():
            raise RuntimeError(
                f"Invalid provider_priority: '{provider_priority}' parsing block {ret}"
            )
        provider_priority = int(provider_priority)
    else:
        provider_priority = None

    return ApkindexBlock(
        arch=Arch.from_str(ret["arch"]),
        depends=ret["depends"],
        origin=ret.get("origin"),
        pkgname=ret["pkgname"],
        provides=ret["provides"],
        provider_priority=provider_priority,
        timestamp=ret.get("timestamp"),
        version=ret["version"],
    )


@overload
def parse_add_block(
    ret: dict[str, ApkindexBlock],
    block: ApkindexBlock,
    alias: str | None = ...,
    multiple_providers: Literal[False] = ...,
) -> None: ...


@overload
def parse_add_block(
    ret: dict[str, dict[str, ApkindexBlock]],
    block: ApkindexBlock,
    alias: str | None = ...,
    multiple_providers: Literal[True] = ...,
) -> None: ...


@overload
def parse_add_block(
    ret: dict[str, ApkindexBlock] | dict[str, dict[str, ApkindexBlock]],
    block: ApkindexBlock,
    alias: str | None = ...,
    multiple_providers: bool = ...,
) -> None: ...


def parse_add_block(
    ret: dict[str, ApkindexBlock] | dict[str, dict[str, ApkindexBlock]],
    block: ApkindexBlock,
    alias: str | None = None,
    multiple_providers: bool = True,
) -> None:
    """
    Add one block to the return dictionary of parse().

    :param ret: dictionary of all packages in the APKINDEX that is
                getting built right now. This function will extend it.
    :param block: return value from parse_next_block().
    :param alias: defaults to the pkgname, could be an alias from the
                  "provides" list.
    :param multiple_providers: assume that there are more than one provider for
                               the alias. This makes sense when parsing the
                               APKINDEX files from a repository (#1122), but
                               not when parsing apk's installed packages DB.
    """
    # Defaults
    pkgname = block.pkgname
    alias = alias or pkgname

    # Get an existing block with the same alias
    block_old = None
    if multiple_providers:
        ret = cast(dict[str, dict[str, ApkindexBlock]], ret)
        if alias in ret and pkgname in ret[alias]:
            picked_aliases = ret[alias]
            if not isinstance(picked_aliases, dict):
                raise AssertionError
            block_old = picked_aliases[pkgname]
    elif not multiple_providers:
        if alias in ret:
            ret = cast(dict[str, ApkindexBlock], ret)
            picked_alias = ret[alias]
            if not isinstance(picked_alias, ApkindexBlock):
                raise AssertionError
            block_old = picked_alias

    # Ignore the block, if the block we already have has a higher version
    if block_old:
        version_old = block_old.version
        version_new = block.version
        if pmb.parse.version.compare(version_old, version_new) == 1:
            return

    # Add it to the result set
    if multiple_providers:
        ret = cast(dict[str, dict[str, ApkindexBlock]], ret)
        if alias not in ret:
            ret[alias] = {}
        picked_aliases = cast(dict[str, ApkindexBlock], ret[alias])
        picked_aliases[pkgname] = block
    else:
        ret = cast(dict[str, ApkindexBlock], ret)
        ret[alias] = block


@overload
def parse(path: Path) -> dict[str, dict[str, ApkindexBlock]]: ...


@overload
def parse(path: Path, multiple_providers: Literal[False] = ...) -> dict[str, ApkindexBlock]: ...


@overload
def parse(
    path: Path, multiple_providers: Literal[True] = ...
) -> dict[str, dict[str, ApkindexBlock]]: ...


def parse(
    path: Path, multiple_providers: bool = True
) -> dict[str, ApkindexBlock] | dict[str, dict[str, ApkindexBlock]]:
    r"""
    Parse an APKINDEX.tar.gz file, and return its content as dictionary.

    :param path: path to an APKINDEX.tar.gz file or apk package database
                 (almost the same format, but not compressed).
    :param multiple_providers: assume that there are more than one provider for
                               the alias. This makes sense when parsing the
                               APKINDEX files from a repository (#1122), but
                               not when parsing apk's installed packages DB.
    :returns: (without multiple_providers)

    Generic format:
        ``{ pkgname: ApkindexBlock, ... }``

    Example:
        ``{ "postmarketos-mkinitfs": ApkindexBlock, "so:libGL.so.1": ApkindexBlock, ...}``

    :returns: (with multiple_providers)

    Generic format:
        ``{ provide: { pkgname: ApkindexBlock, ... }, ... }``

    Example:
        ``{ "postmarketos-mkinitfs": {"postmarketos-mkinitfs": ApkindexBlock},"so:libGL.so.1": {"mesa-egl": ApkindexBlock, "libhybris": ApkindexBlock}, ...}``

    *NOTE:* ``block`` is the return value from ``parse_next_block()`` above.

    """
    # Require the file to exist
    if not path.is_file():
        logging.verbose(
            "NOTE: APKINDEX not found, assuming no binary packages"
            f" exist for that architecture: {path}"
        )
        return {}

    # Try to get a cached result first
    lastmod = path.lstat().st_mtime
    cache_key_ = "multiple" if multiple_providers else "single"
    key = cache_key(path)
    if key in pmb.helpers.other.cache["apkindex"]:
        cache = pmb.helpers.other.cache["apkindex"][key]
        if cache["lastmod"] == lastmod:
            if cache_key_ in cache:
                return cache[cache_key_]
        else:
            clear_cache(path)

    # Read all lines
    lines: Sequence[str]
    if tarfile.is_tarfile(path):
        with (
            tarfile.open(path, "r:gz") as tar,
            tar.extractfile(tar.getmember("APKINDEX")) as handle,  # type:ignore[union-attr]
        ):
            lines = handle.read().decode().splitlines()
    else:
        with path.open("r", encoding="utf-8") as handle:
            lines = handle.read().splitlines()

    # The APKINDEX might be empty, for example if you run "pmbootstrap index" and have no local
    # packages
    if not lines:
        return {}

    # Parse the whole APKINDEX file
    ret: dict[str, ApkindexBlock] = collections.OrderedDict()
    if lines[-1] == "\n":
        lines.pop()  # Strip the trailing newline
    while True:
        block = parse_next_block(path, lines)
        if not block:
            break

        # Skip virtual packages
        if block.timestamp is None:
            logging.verbose(f"Skipped virtual package {block} in file: {path}")
            continue

        # Add the next package and all aliases
        parse_add_block(ret, block, None, multiple_providers)
        for alias in block.provides:
            parse_add_block(ret, block, alias, multiple_providers)

    # Update the cache
    key = cache_key(path)
    if key not in pmb.helpers.other.cache["apkindex"]:
        pmb.helpers.other.cache["apkindex"][key] = {"lastmod": lastmod}
    pmb.helpers.other.cache["apkindex"][key][cache_key_] = ret
    return ret


def parse_blocks(path: Path) -> list[ApkindexBlock]:
    """
    Read all blocks from an APKINDEX.tar.gz into a list.

    :path: full path to the APKINDEX.tar.gz file.
    :returns: all blocks in the APKINDEX, without restructuring them by
              pkgname or removing duplicates with lower versions (use
              parse() if you need these features).
    """
    # Parse all lines
    with tarfile.open(path, "r:gz") as tar, tar.extractfile(tar.getmember("APKINDEX")) as handle:  # type:ignore[union-attr]
        lines = handle.read().decode().splitlines()

    # Parse lines into blocks
    ret: list[ApkindexBlock] = []
    while True:
        block = pmb.parse.apkindex.parse_next_block(path, lines)
        if not block:
            return ret
        ret.append(block)


# FIXME: come up with something better here...
def cache_key(path: Path) -> int:
    return hash(path)


def clear_cache(path: Path) -> bool:
    """
    Clear the APKINDEX parsing cache.

    :returns: True on successful deletion, False otherwise
    """
    key = cache_key(path)
    logging.verbose(f"Clear APKINDEX cache for: {key}")
    if key in pmb.helpers.other.cache["apkindex"]:
        del pmb.helpers.other.cache["apkindex"][key]
        return True
    else:
        logging.verbose(
            "Nothing to do, path was not in cache:"
            + str(pmb.helpers.other.cache["apkindex"].keys())
        )
        return False


def providers(
    package: str,
    arch: Arch | None = None,
    must_exist: bool = True,
    indexes: list[Path] | None = None,
    user_repository: bool = True,
) -> dict[str, ApkindexBlock]:
    """
    Get all packages, which provide one package.

    :param package: of which you want to have the providers
    :param arch: defaults to native arch, only relevant for indexes=None
    :param must_exist: When set to true, raise an exception when the package is
                       not provided at all.
    :param indexes: list of APKINDEX.tar.gz paths, defaults to all index files
                    (depending on arch)
    :param user_repository: add path to index of locally built packages
    :returns: list of parsed packages. Example for package="so:libGL.so.1":
        ``{"mesa-egl": ApkindexBlock, "libhybris": ApkindexBlock}``
    """
    if not indexes:
        indexes = pmb.helpers.repo.apkindex_files(arch, user_repository=user_repository)

    pkgname_with_op = package
    package = pmb.helpers.package.remove_operators(pkgname_with_op)

    ret: dict[str, ApkindexBlock] = collections.OrderedDict()
    for path in indexes:
        # Skip indexes not providing the package
        index_packages = parse(path)
        if package not in index_packages:
            continue

        indexed_package = index_packages[package]

        # Iterate over found providers
        for provider_pkgname, provider in indexed_package.items():
            version = provider.version
            if not pmb.helpers.package.check_version_constraints(pkgname_with_op, version):
                continue
            # Skip lower versions of providers we already found
            if provider_pkgname in ret:
                version_last = ret[provider_pkgname].version
                if pmb.parse.version.compare(version, version_last) == -1:
                    logging.verbose(
                        f"{package}: provided by: {provider_pkgname}-{version}"
                        f"in {path} (but {version_last} is higher)"
                    )
                    continue

            # Add the provider to ret
            logging.verbose(f"{package}: provided by: {provider_pkgname}-{version} in {path}")
            ret[provider_pkgname] = provider

    if ret == {} and must_exist:
        import os

        logging.debug(f"Searched in APKINDEX files: {', '.join([os.fspath(x) for x in indexes])}")
        raise RuntimeError("Could not find package '" + package + "'!")

    return ret


def provider_highest_priority(
    providers: dict[str, ApkindexBlock], pkgname: str
) -> dict[str, ApkindexBlock]:
    """
    Get the provider(s) with the highest provider_priority and log a message.

    :param providers: returned dict from providers(), must not be empty
    :param pkgname: the package name we are interested in (for the log message)
    """
    max_priority = 0
    priority_providers: collections.OrderedDict[str, ApkindexBlock] = collections.OrderedDict()
    for provider_name, provider in providers.items():
        priority = int(-1 if provider.provider_priority is None else provider.provider_priority)
        if priority > max_priority:
            priority_providers.clear()
            max_priority = priority
        if priority == max_priority:
            priority_providers[provider_name] = provider

    if priority_providers:
        logging.debug(
            f"{pkgname}: picked provider(s) with highest priority "
            f"{max_priority}: {', '.join(priority_providers.keys())}"
        )
        return priority_providers

    # None of the providers seems to have a provider_priority defined
    return providers


def provider_shortest(providers: dict[str, ApkindexBlock], pkgname: str) -> ApkindexBlock:
    """
    Get the provider with the shortest pkgname and log a message. In most cases
    this should be sufficient, e.g. 'mesa-purism-gc7000-egl, mesa-egl' or
    'gtk+2.0-maemo, gtk+2.0'.

    :param providers: returned dict from providers(), must not be empty
    :param pkgname: the package name we are interested in (for the log message)
    """
    ret = min(list(providers.keys()), key=len)
    if len(providers) != 1:
        logging.debug(
            f"{pkgname}: has multiple providers ("
            f"{', '.join(providers.keys())}), picked shortest: {ret}"
        )
    return providers[ret]


# This can't be cached because the APKINDEX can change during pmbootstrap build!
def package(
    package: str,
    arch: Arch | None = None,
    must_exist: bool = True,
    indexes: list[Path] | None = None,
    user_repository: bool = True,
) -> ApkindexBlock | None:
    """
    Get a specific package's data from an apkindex.

    :param package: of which you want to have the apkindex data
    :param arch: defaults to native arch, only relevant for indexes=None
    :param must_exist: When set to true, raise an exception when the package is
                       not provided at all.
    :param indexes: list of APKINDEX.tar.gz paths, defaults to all index files
                    (depending on arch)
    :param user_repository: add path to index of locally built packages
    :returns: ApkindexBlock or None when the package was not found.
    """
    # Provider with the same package
    package_providers = providers(
        package, arch, must_exist, indexes, user_repository=user_repository
    )

    if package_providers:
        providers_priority = provider_highest_priority(package_providers, package)
        num_providers = len(providers_priority)
        # Only one provider with max priority: return it
        if num_providers == 1:
            return next(iter(providers_priority.values()))
        # Multiple providers with max priority: let follow-on logic decide
        elif num_providers > 1:
            package_providers = providers_priority

    # Any provider
    if package_providers:
        return pmb.parse.apkindex.provider_shortest(package_providers, package)

    # No provider
    if must_exist:
        raise RuntimeError("Package '" + package + "' not found in any APKINDEX.")
    return None



# Copyright 2026 Stefan Hansson, Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import os
from pathlib import Path

import pmb.chroot.initfs
import pmb.helpers.run
from pmb.core import Chroot, ChrootType
from pmb.core.context import get_context
from pmb.helpers import logging


def odin(device: str, folder: Path) -> None:
    """
    Create Odin flashable tar file with kernel and initramfs
    for devices configured with the flasher method 'heimdall-isorec'
    and with boot.img for devices with 'heimdall-bootimg'
    """
    pmb.flasher.init(device, "heimdall-isorec")
    suffix = Chroot(ChrootType.ROOTFS, device)
    deviceinfo = pmb.parse.deviceinfo(device)

    # Validate method
    method = deviceinfo.flash_method or ""
    if not method.startswith("heimdall-"):
        raise RuntimeError(
            "An odin flashable tar is not supported"
            f" for the flash method '{method}' specified"
            " in the current configuration."
            " Only 'heimdall' methods are supported."
        )

    # Partitions
    partition_kernel = deviceinfo.flash_heimdall_partition_kernel or "KERNEL"
    partition_initfs = deviceinfo.flash_heimdall_partition_initfs or "RECOVERY"

    # Temporary folder
    temp_folder = "/tmp/odin-flashable-tar"
    if (Chroot.native() / temp_folder).exists():
        pmb.chroot.root(["rm", "-rf", temp_folder])

    # Odin flashable tar generation script
    # (because redirecting stdin/stdout is not allowed
    # in pmbootstrap's chroot/shell functions for security reasons)
    odin_script = Chroot(ChrootType.ROOTFS, device) / "tmp/_odin.sh"
    with odin_script.open("w") as handle:
        odin_kernel_md5 = f"{partition_kernel}.bin.md5"
        odin_initfs_md5 = f"{partition_initfs}.bin.md5"
        odin_device_tar = f"{device}.tar"
        odin_device_tar_md5 = f"{device}.tar.md5"

        handle.write(f"#!/bin/sh\ncd {temp_folder}\n")
        if method == "heimdall-isorec":
            handle.write(
                # Kernel: copy and append md5
                f"cp /boot/vmlinuz {odin_kernel_md5}\n"
                f"md5sum -t {odin_kernel_md5} >> {odin_kernel_md5}\n"
                # Initramfs: recompress with lzop, append md5
                f"gunzip -c /boot/initramfs"
                f" | lzop > {odin_initfs_md5}\n"
                f"md5sum -t {odin_initfs_md5} >> {odin_initfs_md5}\n"
            )
        elif method == "heimdall-bootimg":
            handle.write(
                # boot.img: copy and append md5
                f"cp /boot/boot.img {odin_kernel_md5}\n"
                f"md5sum -t {odin_kernel_md5} >> {odin_kernel_md5}\n"
            )
        handle.write(
            # Create tar, remove included files and append md5
            f"tar -c -f {odin_device_tar} *.bin.md5\n"
            "rm *.bin.md5\n"
            f"md5sum -t {odin_device_tar} >> {odin_device_tar}\n"
            f"mv {odin_device_tar} {odin_device_tar_md5}\n"
        )

    commands = [
        ["mkdir", "-p", temp_folder],
        ["cat", "/tmp/_odin.sh"],  # for the log
        ["sh", "/tmp/_odin.sh"],
        ["rm", "/tmp/_odin.sh"],
    ]
    for command in commands:
        pmb.chroot.root(command, suffix)

    # Move Odin flashable tar to native chroot and cleanup temp folder
    pmb.chroot.user(["mkdir", "-p", "/home/pmos/rootfs"])
    (
        pmb.chroot.root(
            [
                "mv",
                f"/mnt/rootfs_{device}{temp_folder}/{odin_device_tar_md5}",
                "/home/pmos/rootfs/",
            ]
        ),
    )
    pmb.chroot.root(["chown", "pmos:pmos", f"/home/pmos/rootfs/{odin_device_tar_md5}"])
    pmb.chroot.root(["rmdir", temp_folder], suffix)

    # Create the symlink
    file = Chroot.native() / "home/pmos/rootfs" / odin_device_tar_md5
    link = folder / odin_device_tar_md5
    pmb.helpers.file.symlink(file, link)

    # Display a readable message
    msg = f" * {odin_device_tar_md5}"
    if method == "heimdall-isorec":
        msg += " (Odin flashable file, contains initramfs and kernel)"
    elif method == "heimdall-bootimg":
        msg += " (Odin flashable file, contains boot.img)"
    logging.info(msg)


def symlinks(target: Path) -> None:
    """Create convenience symlinks to the rootfs and boot files."""
    device = get_context().config.device
    arch = pmb.parse.deviceinfo(device).arch

    # File descriptions
    info = {
        "boot.img": (
            "Fastboot compatible boot.img file, contains initramfs with bootimg header <=2 and kernel"
        ),
        "vendor_boot.img": (
            "Fastboot compatible vendor_boot.img file, contains cmdline, initramfs and dtb. Only with bootimg header >= v3"
        ),
        "dtbo.img": "Fastboot compatible dtbo image",
        "initramfs": "Initramfs",
        "initramfs-extra": "Extra initramfs files in /boot",
        "uInitrd": "Initramfs, legacy u-boot image format",
        "uImage": "Kernel, legacy u-boot image format",
        "vmlinuz": "Linux kernel",
        f"{device}.img": "Rootfs with partitions for /boot and /",
        f"{device}-boot.img": "Boot partition image",
        f"{device}-root.img": "Root partition image",
        f"pmos-{device}.zip": "Android recovery flashable zip",
        "lk2nd.img": "Secondary Android bootloader",
    }

    # Generate a list of patterns
    chroot_native = Chroot.native()
    path_boot = Chroot(ChrootType.ROOTFS, device) / "boot"
    chroot_buildroot = Chroot.buildroot(arch)
    files: list[Path] = [
        path_boot / "boot.img",
        path_boot / "vendor_boot.img",
        path_boot / "uInitrd",
        path_boot / "uImage",
        path_boot / "dtbo.img",
        chroot_native / "home/pmos/rootfs" / f"{device}.img",
        chroot_native / "home/pmos/rootfs" / f"{device}-boot.img",
        chroot_native / "home/pmos/rootfs" / f"{device}-root.img",
        chroot_buildroot / "var/lib/postmarketos-android-recovery-installer" / f"pmos-{device}.zip",
        path_boot / "lk2nd.img",
    ]

    files += list(path_boot.glob("initramfs*"))
    files += list(path_boot.glob("vmlinuz*"))

    # Iterate through all files
    for file in files:
        basename = file.name
        link = target / basename

        # Display a readable message
        msg = " * " + basename
        if basename in info:
            msg += " (" + info[basename] + ")"
        logging.info(msg)

        pmb.helpers.file.symlink(file, link)

    # Create dtbs in folder
    dtbs = list(path_boot.glob("*.dtb"))
    if len(dtbs) == 0:
        return

    logging.info(" * dtbs/ (Device tree blobs)")
    target = target / "dtbs"
    target.mkdir(exist_ok=True)
    for dtb in dtbs:
        pmb.helpers.file.symlink(dtb, target / dtb.name)


def export(target: Path, autoinstall: bool, odin_flashable_tar: bool) -> None:
    config = get_context().config
    # Create the export folder
    if not os.path.exists(target):
        pmb.helpers.run.user(["mkdir", "-p", target])

    # Rootfs image note
    chroot = Chroot.native()
    rootfs_dir = chroot / "home/pmos/rootfs"
    if not any(rootfs_dir.glob(f"{config.device}*.img")):
        logging.info(
            "NOTE: To export the rootfs image, run 'pmbootstrap"
            " install' first (without the 'disk' parameter)."
        )

    # Rebuild the initramfs, just to make sure (see #69)
    if autoinstall:
        pmb.chroot.initfs.build(Chroot(ChrootType.ROOTFS, config.device))

    # Do the export, print all files
    logging.info(f"Export symlinks to: {target}")
    if odin_flashable_tar:
        odin(config.device, target)
    symlinks(target)



# Copyright 2023 Oliver Smith
# SPDX-License-Identifier: GPL-3.0-or-later
import os
from pathlib import Path

import pmb.aportgen.busybox_static
import pmb.aportgen.core
import pmb.aportgen.device
import pmb.aportgen.gcc
import pmb.aportgen.grub_efi
import pmb.aportgen.linux
import pmb.aportgen.musl
import pmb.config
import pmb.helpers.cli
from pmb.core.context import get_context
from pmb.core.pkgrepo import pkgrepo_default_path
from pmb.helpers import logging
from pmb.helpers.exceptions import NonBugError


def get_cross_package_arches(pkgname: str) -> str:
    """
    Get the arches for which we want to build cross packages.

    :param pkgname: package name, e.g. "gcc-aarch64", "gcc-x86_64"

    :returns: string of architecture(s) (space separated). It doesn't
              necessarily make sense to use Arch here given that this value gets
              used to write APKBUILD files, where the ``arch`` field can have values
              that aren't necessarily valid arches like "!armhf", "noarch", or
              "x86 x86_64".
    """
    if pkgname.endswith("-x86_64"):
        return "aarch64"
    else:
        return "x86_64"


def properties(
    pkgname: str,
    folder: Path | None = None,
    device_category: pmb.helpers.devices.DeviceCategory | None = None,
) -> tuple[str, Path, bool]:
    """
    Get the properties for the aport generator, based on the pkgname prefix.

    Example: "musl-armhf" => ("musl", "cross", {"confirm_overwrite": False})

    :param pkgname: package name
    :param folder: optional base folder override
    :param device_category: optional device category for device/linux aports

    :returns: (prefix, folder, confirm_overwrite)
    """
    for prefix in ["busybox-static", "gcc", "musl", "grub-efi"]:
        if pkgname.startswith(prefix):
            return (prefix, folder or Path("cross"), False)

    for prefix in ["device", "linux"]:
        if pkgname.startswith(prefix):
            if not folder:
                folder = (
                    Path("device") / str(device_category) if device_category else Path("device")
                )
            return (prefix, folder, True)

    logging.info(
        "NOTE: aportgen is for generating postmarketOS specific aports, such as the cross-compiler "
        "related packages or the linux kernel fork packages."
    )
    logging.info(
        "NOTE: If you wanted to package new software in general, try"
        " 'pmbootstrap newapkbuild' to generate a template."
    )
    raise NonBugError(f"No generator available for {pkgname}!")


def generate(
    pkgname: str,
    fork_alpine: bool = False,
    fork_alpine_retain_branch: bool = False,
    folder: Path | None = None,
    device_category: pmb.helpers.devices.DeviceCategory | None = None,
) -> None:
    if pkgname.startswith(("device", "linux")) and not device_category:
        device_category = pmb.config.ask_for_mainline_downstream()

    if fork_alpine:
        prefix, folder, confirm_overwrite = (pkgname, Path("temp"), True)
    else:
        prefix, folder, confirm_overwrite = properties(pkgname, folder, device_category)
    config = get_context().config
    path_target = pkgrepo_default_path() / folder / pkgname

    # Confirm overwrite
    if confirm_overwrite and os.path.exists(path_target):
        logging.warning(f"WARNING: Target folder already exists: {path_target}")
        if not pmb.helpers.cli.confirm("Continue and overwrite?"):
            raise NonBugError("Aborted.")

    aportgen = config.work / "aportgen"

    if os.path.exists(aportgen):
        pmb.helpers.run.user(["rm", "-r", aportgen])
    if fork_alpine:
        upstream = pmb.aportgen.core.get_upstream_aport(
            pkgname, retain_branch=fork_alpine_retain_branch
        )
        pmb.helpers.run.user(["cp", "-r", upstream, aportgen])
        pmb.aportgen.core.rewrite(pkgname)
    else:
        match prefix:
            case "busybox-static":
                pmb.aportgen.busybox_static.generate(pkgname)
            case "device":
                # Ignore mypy 'error: Argument 2 to "generate" has incompatible type
                # "DeviceCategory | None"; expected "DeviceCategory".
                # The check on the top of the page already ensures device_category is not
                # None in this case.
                pmb.aportgen.device.generate(pkgname, device_category)  # type: ignore[arg-type]
            case "gcc":
                pmb.aportgen.gcc.generate(pkgname)
            case "grub-efi":
                pmb.aportgen.grub_efi.generate(pkgname)
            case "linux":
                # Ignore mypy error; see note for "device" case above.
                pmb.aportgen.linux.generate(pkgname, device_category)  # type: ignore[arg-type]
            case "musl":
                pmb.aportgen.musl.generate(pkgname)
            case _:
                raise ValueError(f"Unexpected prefix {prefix}")

    # Move to the aports folder
    if os.path.exists(path_target):
        pmb.helpers.run.user(["rm", "-r", path_target])
    pmb.helpers.run.user(["mv", aportgen, path_target])

    logging.info(f"*** pmaport generated: {path_target}")



# Copyright 2024 Anri Dellal
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Final

from pmb.helpers import logging

"""
Helper used to configure locale-related settings in pmOS installation.

Locale and layout information is taken from:
* https://en.wikipedia.org/wiki/List_of_ISO_639_language_codes
* https://en.wikipedia.org/wiki/List_of_ISO_3166_country_codes
* Xkb source code: /usr/share/X11/xkb/rules/base.lst
"""


class XkbLayout:
    primary_layouts: Final[frozenset[str]] = frozenset(
        [
            "at",
            "au",
            "br",
            "ca",
            "ch",
            "de",
            "dk",
            "ee",
            "es",
            "fi",
            "fo",
            "fr",
            "gb",
            "ie",
            "is",
            "it",
            "latam",
            "lt",
            "mt",
            "nz",
            "pl",
            "pt",
            "ro",
            "se",
            "sk",
            "tr",
            "us",
            "vn",
        ]
    )

    def __init__(self, layout: str = "us", variant: str = "") -> None:
        self.layout = layout
        self.variant = variant
        self.options = ""
        self.layout_list: list[str] = []
        self.variant_list: list[str] = []
        if layout in XkbLayout.primary_layouts:
            self.layout_list.append(layout)
            if variant:
                self.variant_list.append(variant)
        elif not self.is_default():
            self.layout_list.extend(["us", layout])
            if variant:
                self.variant_list.extend(["", variant])
            self.options = "grp:alt_shift_toggle"

    def is_default(self) -> bool:
        return self.layout == "us" and not self.variant

    def get_profile_vars(self) -> str | None:
        if self.is_default():
            return None
        template = "XKB_DEFAULT_LAYOUT=" + ",".join(self.layout_list)
        if self.variant_list:
            template += "\nXKB_DEFAULT_VARIANT=" + ",".join(self.variant_list)
        if self.options:
            template += "\nXKB_DEFAULT_OPTIONS=" + self.options
        return template

    def get_keyboard_config(self) -> str | None:
        if self.is_default() or not self.layout_list:
            return None
        template = f"""Section "InputClass"
\tIdentifier "keyboard"
\tMatchIsKeyboard "on"
\tOption "XkbLayout" "{",".join(self.layout_list)}"\n"""
        if self.variant_list:
            template += f'\tOption "XkbVariants" "{",".join(self.variant_list)}"\n'
        if self.options:
            template += f'\tOption "XkbOptions" "{self.options}"\n'
        template += "EndSection"
        return template

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, XkbLayout):
            return False
        return self.layout == other.layout and self.variant == other.variant


def get_xkb_layout(locale: str) -> XkbLayout:
    """
    Get Xkb layout for the given locale.

    :param locale: Locale to get XkbLayout for.
    :return: XkbLayout for the given locale.
    """
    try:
        (language, territory) = locale.split("_")
    except ValueError:
        logging.warning(f'Keyboard config for locale "{locale}" cannot be provided')
        return XkbLayout()

    if language in lang_layouts:
        return XkbLayout(language)
    elif language in lang_to_layout:
        return XkbLayout(lang_to_layout[language])

    match language:
        case "af":
            return XkbLayout("za")
        case "ak":
            return XkbLayout("gh", "akan")
        case "ar":
            if territory == "DZ":
                return XkbLayout("dz", "ar")
            elif territory == "EG":
                return XkbLayout("eg")
            elif territory == "IQ":
                return XkbLayout("iq")
            elif territory == "MA":
                return XkbLayout("ma")
            elif territory == "SY":
                return XkbLayout("sy")
            else:
                return XkbLayout("ara")
        case "as":
            return XkbLayout("in", "asm-kagapa")
        case "ast":
            return XkbLayout("es", "ast")
        case "ber":
            if territory == "DZ":
                return XkbLayout("dz")
            else:  # "MA"
                return XkbLayout("ma")
        case "bo":
            return XkbLayout("cn", "tib")
        case "ca":
            return XkbLayout("es", "cat")
        case "chr":
            return XkbLayout("us", "chr")
        case "crh":
            return XkbLayout("ua", "crh")
        case "de":
            if territory == "AT":
                return XkbLayout("at")
            elif territory == "CH":
                return XkbLayout("ch")
            else:
                return XkbLayout("de")
        case "en":
            if territory == "AU":
                return XkbLayout("au")
            elif territory == "CA":
                return XkbLayout("ca", "eng")
            elif territory == "GB":
                return XkbLayout("gb")
            elif territory == "IN":
                return XkbLayout("in", "eng")
            elif territory == "NG":
                return XkbLayout("ng")
            elif territory == "NZ":
                return XkbLayout("nz")
            elif territory == "ZA":
                return XkbLayout("za")
            else:
                return XkbLayout("us")
        case "es":
            if territory in latam_es_territories:
                return XkbLayout("latam")
            else:  # "ES", "US"
                return XkbLayout("es")
        case "fr":
            if territory == "CA":
                return XkbLayout("ca")
            elif territory == "CH":
                return XkbLayout("ch", "fr")
            else:  # "FR"
                return XkbLayout("fr")
        case "fur":
            return XkbLayout("it", "fur")
        case "gu":
            return XkbLayout("in", "guj")
        case "gv":
            return XkbLayout("gb", "gla")
        case "ha":
            return XkbLayout("ng", "hausa")
        case "hif" | "hne":
            return XkbLayout("in")
        case "ig":
            return XkbLayout("ng", "igbo")
        case "iu":
            return XkbLayout("ca", "ike")
        case "kab":
            return XkbLayout("dz", "azerty-deadkeys")
        case "kn":
            return XkbLayout("in", "kan")
        case "ku":
            return XkbLayout("tr", "ku")
        case "mhr":
            return XkbLayout("ru", "chm")
        case "ml":
            return XkbLayout("in", "mal")
        case "mni":
            return XkbLayout("in", "mni")
        case "mnw":
            return XkbLayout("mm", "mnw")
        case "mr":
            return XkbLayout("in", "marathi")
        case "nan":
            return XkbLayout("cn")  # Min Nan Chinese
        case "nds":
            return XkbLayout("de")  # Low German
        case "oc":
            return XkbLayout("fr", "oci")
        case "or":
            return XkbLayout("in", "ori")
        case "os":
            return XkbLayout("ru", "os_winkeys")
        case "pa":
            return XkbLayout("in", "guru")
        case "ps":
            return XkbLayout("af", "ps")
        case "pt":
            if territory == "BR":
                return XkbLayout("br")
            else:  # if territory == "PT":
                return XkbLayout("pt")
        case "sa":
            return XkbLayout("in", "sas")
        case "sah":
            return XkbLayout("ru", "sah")
        case "sat":
            return XkbLayout("in", "sat")
        case "sd":
            return XkbLayout("in", "sdh")
        case "se":
            return XkbLayout("no", "smi")
        case "sgs":
            return XkbLayout("lt", "sgs")
        case "shn":
            return XkbLayout("mm", "shn")
        case "si":
            return XkbLayout("pk", "snd")
        case "sw":
            if territory == "KE":
                return XkbLayout("ke")
            else:  # "TZ"
                return XkbLayout("tz")
        case "szl":
            return XkbLayout("pl", "szl")
        case "ta":
            if territory == "LK":
                return XkbLayout("lk", "tam_unicode")
            else:  # "IN"
                return XkbLayout("in", "tamilnet")
        case "te":
            return XkbLayout("in", "tel")
        case "tt":
            return XkbLayout("ru", "tt")
        case "ug":
            return XkbLayout("cn", "ug")
        case "ur":
            return XkbLayout("pak", "urd-nla")
        case "yo":
            return XkbLayout("ng", "yoruba")
        case _:
            pass
    if language not in locales_without_layout:
        logging.warning(f'Language "{language}" not found in language list')
    return XkbLayout()  # return default layout if no layout was found


# Layouts which have same code as language
lang_layouts = [
    "az",
    "bg",
    "csb",
    "cv",
    "dsb",
    "fi",
    "fo",
    "hr",
    "hu",
    "id",
    "is",
    "it",
    "lt",
    "lv",
    "mk",
    "mn",
    "mt",
    "nl",
    "pl",
    "ro",
    "ru",
    "sk",
    "tg",
    "th",
    "tr",
    "uk",
    "uz",
]
lang_to_layout = {
    "am": "et",
    "be": "by",
    "bn": "bd",
    "br": "bre",
    "bs": "ba",
    "cmn": "cn",
    "cs": "cz",
    "da": "dk",
    "dz": "bt",
    "el": "gr",
    "eo": "epo",
    "et": "ee",
    "eu": "es",
    "fa": "ir",
    "fil": "ph",
    "ga": "ie",
    "gd": "gla",
    "hak": "cn",
    "he": "il",
    "hi": "in",
    "hy": "am",
    "ja": "jp",
    "ka": "ge",
    "kk": "kz",
    "km": "kh",
    "ko": "kr",
    "ky": "kg",
    "lo": "la",
    "lzh": "cn",
    "mai": "in",
    "mi": "mao",
    "ms": "my",
    "my": "mm",
    "nb": "no",
    "ne": "np",
    "nn": "no",
    "sl": "si",
    "sq": "al",
    "sr": "rs",
    "sv": "se",
    "tk": "tm",
    "tl": "in",
    "tn": "bw",
    "vi": "vn",
    "wo": "sn",
    "yi": "il",
    "yue": "cn",
    "zh": "cn",
}
# Locales which either have no layout or missing information about one
locales_without_layout = [
    "aa",  # Afar
    "agr",  # Aguaruna
    "an",  # Aragonese
    "anp",  # Angika
    "ayc",  # Aymara
    "bem",  # Bemba
    "bhb",  # Bhili
    "bho",  # Bhojpuri
    "bi",  # Bislama
    "bn",  # Bengali
    "brx",  # Bodo (India)
    "byn",  # Bilin
    "ce",  # Chechen
    "ch",  # Chamorro
    "cy",  # Wales
    "doi",  # Dogri
    "dv",  # Divehi
    "ff",  # Fulah
    "fy",  # Western Frisian
    "gez",  # Geez
    "gl",  # Galician
    "hsb",  # Upper Sorbian
    "ht",  # Haitian
    "ia",  # Interlingua
    "ik",  # Inupiaq
    "kl",  # Kalaallisut, Greenlandic
    "kok",  # Konkani
    "ks",  # Kashmiri
    "kw",  # Kinyarwanda
    "lb",  # Luxembourgish
    "lg",  # Ganda
    "li",  # Limburgish
    "lij",  # Ligurian
    "ln",  # Lingala
    "mfe",  # Morisyen
    "mg",  # Malagasy
    "miq",  # Miskito
    "mjw",  # Karbi
    "nhn",  # Nahuatl
    "niu",  # Niuean
    "nr",  # South Ndebele
    "nso",  # Northern Sotho
    "om",  # Oromo
    "pap",  # Papiamento
    "quz",  # Cusco Quechua
    "raj",  # Rajasthani
    "rw",  # Kinyarwanda
    "sc",  # Sardinian
    "shs",  # Shuswap
    "sid",  # Sidamo
    "sm",  # Samoan
    "so",  # Somali
    "ss",  # Swati
    "st",  # Southern Sotho
    "tcy",  # Tulu
    "the",  # Chitwania Tharu
    "ti",  # Tigrinya
    "tig",  # Tigre
    "to",  # Tonga
    "tpi",  # Tok Pisin
    "ts",  # Tsonga
    "unm",  # Unami
    "ve",  # Venda
    "wa",  # Walloon
    "wae",  # Walser
    "wal",  # Wolaitta
    "xh",  # Xhosa
    "yuw",  # Yau (Morobe province)
    "zu",  # Zulu
]
latam_es_territories = [
    "AR",
    "BO",
    "CL",
    "CO",
    "CR",
    "CU",
    "DO",
    "EC",
    "GT",
    "HN",
    "MX",
    "NI",
    "PA",
    "PE",
    "PR",
    "PY",
    "SV",
    "UY",
    "VE",
]



# Copyright 2023 Attila Szollosi
# SPDX-License-Identifier: GPL-3.0-or-later
from pathlib import Path

import pmb.chroot
import pmb.chroot.apk
import pmb.flasher
import pmb.helpers.frontend
from pmb.core.chroot import Chroot
from pmb.helpers import logging
from pmb.types import PmbArgs


def create_zip(args: PmbArgs, chroot: Chroot, device: str) -> None:
    """Create android recovery compatible installer zip."""
    zip_root = Path("/var/lib/postmarketos-android-recovery-installer/")
    rootfs = "/mnt/rootfs_" + device
    flavor = pmb.helpers.frontend._parse_flavor(device)
    deviceinfo = pmb.parse.deviceinfo()
    method = deviceinfo.flash_method
    fvars = pmb.flasher.variables(
        flavor,
        method,
        getattr(args, "cmdline", None),
        getattr(args, "no_reboot", None),
        getattr(args, "partition", None),
        getattr(args, "resume", None),
    )

    # Install recovery installer package in buildroot
    pmb.chroot.apk.install(["postmarketos-android-recovery-installer"], chroot)

    logging.info(f"({chroot}) create recovery zip")

    for key in fvars:
        fvalue = fvars[key]

        if fvalue is None:
            continue

        pmb.flasher.check_partition_blacklist(deviceinfo, key, fvalue)

    if (
        fvars["$PARTITION_KERNEL"] is None
        or fvars["$PARTITION_INITFS"] is None
        or fvars["$PARTITION_ROOTFS"] is None
    ):
        raise AssertionError("Partitions should not be None at this point")

    # Create config file for the recovery installer
    options: dict[str, bool | str] = {
        "DEVICE": device,
        "FLASH_KERNEL": args.recovery_flash_kernel,
        "FLAVOR": "",
        "ISOREC": method == "heimdall-isorec",
        "KERNEL_PARTLABEL": fvars["$PARTITION_KERNEL"],
        "INITFS_PARTLABEL": fvars["$PARTITION_INITFS"],
        # Name is still "SYSTEM", not "ROOTFS" in the recovery installer
        "SYSTEM_PARTLABEL": fvars["$PARTITION_ROOTFS"],
        "INSTALL_PARTITION": args.recovery_install_partition,
        "CIPHER": args.cipher,
        "FDE": args.full_disk_encryption,
    }

    # Write to a temporary file
    config_temp = chroot / "tmp/install_options"
    with config_temp.open("w") as handle:
        for key, value in options.items():
            if isinstance(value, bool):
                value = str(value).lower()
            handle.write(key + "='" + value + "'\n")

    commands = [
        # Move config file from /tmp/ to zip root
        ["mv", "/tmp/install_options", "chroot/install_options"],
        # Create tar archive of the rootfs
        ["tar", "-pcf", "rootfs.tar", "--exclude", "./home", "-C", rootfs, "."],
        # Append packages keys
        ["tar", "-prf", "rootfs.tar", "-C", "/", "./etc/apk/keys"],
        # Compress with -1 for speed improvement
        ["gzip", "-f1", "rootfs.tar"],
        ["build-recovery-zip", device],
    ]
    for command in commands:
        pmb.chroot.root(command, chroot, working_dir=zip_root)



# Copyright 2024 Caleb Connolly
# SPDX-License-Identifier: GPL-3.0-or-later

import copy
import inspect
from collections.abc import Callable
from typing import Any, Generic, TypeVar, overload

FuncArgs = TypeVar("FuncArgs")
FuncReturn = TypeVar("FuncReturn")


class Wrapper(Generic[FuncArgs, FuncReturn]):
    def __init__(self, cache: "Cache", func: Callable[[FuncArgs], FuncReturn]) -> None:
        self.cache = cache
        self.func = func
        self.disabled = False
        self.__module__ = func.__module__
        self.__name__ = func.__name__
        self.hits = 0
        self.misses = 0

    # When someone attempts to call a cached function, they'll
    # actually end up here. We first check if we have a cached
    # result and if not then we do the actual function call and
    # cache it if applicable
    def __call__(self, *args: Any, **kwargs: Any) -> FuncReturn:
        if self.disabled:
            return self.func(*args, **kwargs)

        # Build the cache key from the function arguments that we
        # care about, which might be none of them
        key = self.cache.build_key(self.func, *args, **kwargs)
        # Don't cache
        if key is None:
            self.misses += 1
            return self.func(*args, **kwargs)

        if key not in self.cache.cache:
            self.misses += 1
            self.cache.cache[key] = self.func(*args, **kwargs)
        else:
            self.hits += 1
            if self.cache.cache_deepcopy:
                self.cache.cache[key] = copy.deepcopy(self.cache.cache[key])

        return self.cache.cache[key]

    def cache_clear(self) -> None:
        self.cache.clear()
        self.misses = 0
        self.hits = 0

    def cache_disable(self) -> None:
        self.disabled = True


class Cache:
    """
    Cache decorator for caching function results based on parameters.

    :param args: a list of function arguments to use as the cache key.
    :param kwargs: these are arguments where we should only cache if the
    function is called with the given value. For example, in pmb.build._package
    we never want to use the cached result when called with force=True.
    """

    def __init__(self, *args: str, cache_deepcopy: bool = False, **kwargs: Any) -> None:
        for a in args:
            if not isinstance(a, str):
                raise ValueError(f"Cache key must be a string, not {type(a)}")

        if len(args) != len(set(args)):
            raise ValueError("Duplicate cache key properties")

        self.cache: dict[str, Any] = {}
        self.params = args
        self.kwargs = kwargs
        self.cache_deepcopy = cache_deepcopy

    # Build the cache key, or return None to not cache in the case where
    # we only cache when an argument has a specific value
    def build_key(self, func: Callable, *args: Any, **kwargs: Any) -> str | None:
        key = "~"
        # Easy case: cache irrelevant of arguments
        if not self.params and not self.kwargs:
            return key

        signature = inspect.signature(func)

        passed_args: dict[str, str] = {}
        for i, (k, val) in enumerate(signature.parameters.items()):
            if k in self.params or k in self.kwargs:
                if i < len(args):
                    passed_args[k] = args[i]
                elif k in kwargs:
                    passed_args[k] = kwargs[k]
                elif val.default != inspect.Parameter.empty:
                    passed_args[k] = val.default
                else:
                    raise ValueError(
                        f"Invalid cache key argument {k}"
                        f" in function {func.__module__}.{func.__name__}"
                    )

        for k, v in self.kwargs.items():
            if k not in signature.parameters:
                raise ValueError(
                    f"Cache key attribute {k} is not a valid parameter to {func.__name__}()"
                )
            passed_val = passed_args[k]
            if passed_val != v:
                # Don't cache
                return None
            else:
                key += f"{k}=({v})~"

        if self.params:
            for k, v in passed_args.items():
                if k in self.params:
                    key += f"{v}~"

        return key

    @overload
    def __call__(self, func: Callable[..., FuncReturn]) -> Wrapper[None, FuncReturn]: ...

    @overload
    def __call__(self, func: Callable[[FuncArgs], FuncReturn]) -> Wrapper[FuncArgs, FuncReturn]: ...

    def __call__(self, func: Callable[[FuncArgs], FuncReturn]) -> Wrapper[FuncArgs, FuncReturn]:
        argnames = func.__code__.co_varnames
        for a in self.params:
            if a not in argnames:
                raise ValueError(
                    f"Cache key attribute {a} is not a valid parameter to {func.__name__}()"
                )

        # FIXME: Once PEP-695 generics are in we shouldn't need this.
        return Wrapper(self, func)

    def clear(self) -> None:
        self.cache.clear()
